/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealQE65000Driver.h"
#include <fstream>
#define LINUX
#include <iomanip>
#include "api/seabreezeapi/SeaBreezeAPI.h"
#include "api/SeaBreezeWrapper.h"

using namespace std;

namespace oceanoptics {

RealQE65000Driver::RealQE65000Driver(QE65000Driver* owner) : QE65000State(owner) {

}

RealQE65000Driver::~RealQE65000Driver() {

}

void RealQE65000Driver::init() {
	//compulsory to have!
//	m_owner->sendProgressEvent(PROGRESSION_END_DEVICE_CONTAINER);
	owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
}


void RealQE65000Driver::measure() {
	int32 error = 0;

	//Get deviceID necessary to use library  methods
	sbapi_initialize();
	sbapi_probe_devices();
	int32 numberOfDevices = sbapi_get_number_of_device_ids(); //Provides the number of devices that have either been probed or manually specified

	if (numberOfDevices != 0) {
		int64* deviceIds = (int64 *)calloc(numberOfDevices, sizeof(int64));
		int32 devicesNumber = sbapi_get_device_ids(deviceIds, numberOfDevices); //This provides a unique ID of each device that is detected or specified. The IDs are copied into the buffer deviceIds

		for (int32 i = 0; i < devicesNumber; i++) {
			//the first thing we have to do is to open the spectrometer, necessary?
			int64 deviceId = deviceIds[i];
			int32 result = sbapi_open_device(deviceId, &error);
			if (result == 0) {

				setIntegrationTime(deviceId, owner()->integrationTime());
				getFormattedSpectrum(deviceId);
				getWavelength(deviceId);

				owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
			}
			sbapi_close_device(deviceId, &error);
		}

	} else {
		cout << "no spectrometer found" << endl;
		//Do not forget to send end of progression to unblock start
		owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
	}
}


void RealQE65000Driver::stop() {
	owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
}

/**
 * Check the status of the spectrometer: plugged or unplugged
 */
void RealQE65000Driver::readStatus() {
	sbapi_initialize();
	sbapi_probe_devices();
	int32 numberOfDevices = sbapi_get_number_of_device_ids();
	if (numberOfDevices != 0) {
		owner()->status.update(QE65000Driver::PLUGGED);
	} else {
		owner()->status.update(QE65000Driver::UNPLUGGED);
	}


}

void RealQE65000Driver::setIntegrationTime(int64 deviceID, int32 integrationTime) {

	int32 error = 0;
	//need to get spectrometer feature IDs --> it is the same that opening it...
	int32 spectroNumbers = sbapi_get_number_of_spectrometer_features(deviceID, &error);

	if (spectroNumbers != 0) {
		int64* spectrometerIds = (int64 *) calloc(spectroNumbers, sizeof(int64));
		int32 numberOfSpectrometers = sbapi_get_spectrometer_features(deviceID, &error, spectrometerIds, spectroNumbers);

		for (int32 i = 0; i < numberOfSpectrometers; i++) {
			//NOTE: integration time must be bigger than minimum integration time
			int32 minimumIntegrationTime = sbapi_spectrometer_get_minimum_integration_time_micros(deviceID, spectrometerIds[i], &error);

			if (minimumIntegrationTime > integrationTime) {
				integrationTime = minimumIntegrationTime;
			}
			//api: spectrometerSetIntegrationTimeMicros (long deviceID, long featureID, int *errorCode, unsigned long integrationTimeMicros)
			sbapi_spectrometer_set_integration_time_micros(deviceID, spectrometerIds[i], &error, integrationTime);
		}
	}
}

/**
 * Acquire a spectrum. In this mode, auto-nulling should be automatically performed for devices that support it.
 */
void RealQE65000Driver::getFormattedSpectrum(int64 deviceID) {
	int32 error = 0;
	//need to get spectrometer feature IDs --> it is the same than opening it...
	int32 spectroNumbers = sbapi_get_number_of_spectrometer_features(deviceID, &error);

	if (spectroNumbers != 0) {
		int64* spectrometerIds = (int64 *) calloc(spectroNumbers, sizeof(int64));
		int32 numberOfSpectrometers = sbapi_get_spectrometer_features(deviceID,	&error, spectrometerIds, spectroNumbers);

		int32 length;
		float64* doubleBuffer;
		for (int32 i = 0; i < numberOfSpectrometers; i++) {
			length = sbapi_spectrometer_get_formatted_spectrum_length(deviceID, spectrometerIds[i], &error);
			doubleBuffer = (float64 *)calloc(length, sizeof(float64));
			sbapi_spectrometer_get_formatted_spectrum(deviceID, spectrometerIds[i], &error, doubleBuffer, length);
		}

		//Spectrum = Y data
		owner()->yData.set(doubleBuffer);
		owner()->yData.setSize((int64) length);
	}
}

/*
 * Acquire a spectrum. Returns the answer in raw, unformatted bytes. NOT USED
 */
void RealQE65000Driver::getUnFormattedSpectrum(int64 deviceID) {
	int32 error = 0;
	//need to get spectrometer feature IDs --> it is the same than opening it...
	int32 spectroNumbers = sbapi_get_number_of_spectrometer_features(deviceID, &error);

	if (spectroNumbers != 0) {
		int64* spectrometerIds = (int64 *) calloc(spectroNumbers, sizeof(int64));
		int32 numberOfSpectrometers = sbapi_get_spectrometer_features(deviceID, &error, spectrometerIds, spectroNumbers);

		for (int32 i = 0; i < numberOfSpectrometers; i++) {
			int32 length = sbapi_spectrometer_get_unformatted_spectrum_length(deviceID, spectrometerIds[i], &error);
			unsigned char* byteBuffer = (unsigned char *) calloc(length, sizeof(unsigned char));
			sbapi_spectrometer_get_unformatted_spectrum(deviceID, spectrometerIds[i], &error, byteBuffer, length);
		}
	}
}


/*
 * Get Wavelengths or x data
 */
void RealQE65000Driver::getWavelength(int64 deviceID) {
	int32 error = 0;
	//need to get spectrometer feature IDs --> it is the same than opening it...
	int32 spectroNumbers = sbapi_get_number_of_spectrometer_features(deviceID, &error);

	if (spectroNumbers != 0) {
		int64* spectrometerIds = (int64 *) calloc(spectroNumbers, sizeof(int64));
		int32 numberOfSpectrometers = sbapi_get_spectrometer_features(deviceID,	&error, spectrometerIds, spectroNumbers);

		int32 length;
		float64* valuesBuffer;
		for (int32 i = 0; i < numberOfSpectrometers; i++) {
			//The length depends on measuring formatted or unformatted data. UP to now we stick to formatted data
			length = sbapi_spectrometer_get_formatted_spectrum_length(deviceID, spectrometerIds[i], &error);
			valuesBuffer = (float64 *) calloc(length, sizeof(float64));
			sbapi_spectrometer_get_wavelengths(deviceID, spectrometerIds[i], &error, valuesBuffer, length);
		}

		owner()->xData.set(valuesBuffer);
		owner()->xData.setSize((int64) length);
	}
}

/**
 * Manipulate the shutter. It seems to me that this feature is not available
 */
void RealQE65000Driver::setShutter(int64 deviceID) {
	int32 error = 0;
	char opened = 0; //If 0 -> shutter closed, if 1 -> shutter opened
	int32 numberOfShutters = sbapi_get_number_of_shutter_features(deviceID,	&error);

	if (numberOfShutters != 0) {
		int64* shutterIds = (int64 *)calloc(numberOfShutters, sizeof(int64));
		int32 numberOfShuttersWithFeature = sbapi_get_shutter_features(deviceID, &error, shutterIds, numberOfShutters);

		//Iterate over the shutter features
		for (int32 i = 0; i < numberOfShuttersWithFeature; i++) {
			//open shutter
			sbapi_shutter_set_shutter_open(deviceID, shutterIds[i], &error, 1);
			cout << "shutter open!" << endl;

			//close shutter
			sbapi_shutter_set_shutter_open(deviceID, shutterIds[i], &error, 0);
			cout << "shutter close!" << endl;
		}
	}
}

}




